/*
   
   THE DISCLASS PACKAGE

   Function definitions

   Version                    : 1.2
   Last modification          : October 24, 2002

   Documentation and Reference: see MarkovClasses.pdf

*/

/*

  Copyright (C) 1996, 2002 Dietmar Wendt,
                Hans Ekkehard Plesser <hans.plesser@nlh.no>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/* ========================================================================= */
/*                                                                           */
/*                                  Contents                                 */
/*                                                                           */
/* ========================================================================= */

/* This file consists of two parts:

   - Public function definitions:
       dc_clear
       dc_init
       dc_move
       dc_rand
       dc_store
     which constitute the user interface of the logclass package

   - Private function definitions:
       dc_delete
       dc_error
       dc_reorg
      which are internal to logclass and should not be called by the user

*/


/* ========================================================================= */
/*                                                                           */
/*                                 Includes                                  */
/*                                                                           */
/* ========================================================================= */

#include "dc.h"             /* contains all further includes, definition of  */
			    /* preprocessor commands, type definitions and   */
			    /* function prototypes                           */
#include <rand55.h>         /* Knuth's lagged Fibonacci                      */

/* ========================================================================= */
/*                                                                           */
/*          Private section: users should keep their HANDS OFF !!!           */
/*                                                                           */
/* ========================================================================= */

/* moved to dc.1.2.c */

void dc_delete(dc_global *dc, dc_event *ded);

/* Task:
     - remove event from its class, reduce class and total rates
       accordingly

   Arguments:
     - dc         : pointer to the global data structure
     - ded         : ptr to ded of event

   Called by:
     dc_move

   Calls:
     dc->event_moved
*/

/*--------------------------------------------------------------------------*/

void dc_error(int error, char *err_fmt, ... );

/* Task:
     - issues an error or warning message 
     - terminates execution in case of error calling exit(1)
     - if a pointer to a function is passed as argument, this function
       is called before calling exiting
     - a pointer to user data structures can be passed to the latter fct

   Arguments:
     - error : if == 0, issue a message, but do not terminate
                  != 0, exit after printing message

     - err_fmt == "[d][f[v]][f[v]]...[f[v]][e[p]]"
         format string indicating the content of the variable argument list:
	   + one lowercase character per argument, no spaces
	   + [] indicate optional entries, ... repetition
	   + entries must be in the order given above
	     d: name of file to which error message is to be written
	     f: format string to be passed to fprintf, may contain
	        zero or one conversion characters for DOUBLE 
		values, i.e. %___e, %___f or %___g
	     v: DOUBLE value to be used as argument by fprintf
	     e: pointer to user's own error handling function:
	        must return void, may take zero or one ptr as argument
	     p: pointer of type void* to be passed to users error function
                if !=NULL

     - ...  : arguments as specified in err_fmt

   Returns:

   Called by:
     dc_init, dc_rand, dc_reorg, dc_store

   Calls:
     function of type void() if passed as argument

   Remarks:
     - format string and argument list must match; there is
       no error-checking within dc_error.
*/

/*--------------------------------------------------------------------------*/

void dc_reorg(dc_global *dc);

/* Task:
     - reorganizes memory when a class gets overcrowded
     - after deds have been moved in memory, dc->event_moved is used
       to update the ued->ded link
     - recalculates class and total rates to reduce round-off errors

   Arguments:
     - dc : pointer to the global data structure

   Exits with error message, if
     - all deds allocated have been assigned already
     - memory reorganization failed otherwise

   Called by:
     dc_store

   Calls:
     dc->event_moved, dc_error
*/


/* ========================================================================= */
/*                                                                           */
/*                      Public function declarations                         */
/*                                                                           */
/* ========================================================================= */

void dc_clear(dc_global *dc) 

/* - returns memory allocated by dc_init() to OS
     
   Called by: user program
*/

{ 
  free(dc->cbeg->bot);   /* free event descriptors first, then class descs. */
  free(dc->cbeg);        /* i.e. reverse allocation order                   */
}

/*--------------------------------------------------------------------------*/

void dc_init(dc_global *dc, size_t num_deds, int num_classes,
             double *class_r_ind, int nullclass, long int rand_seed,
             void (*event_moved)(), char err_file[], void (*err_proc),
	     void *err_ptr)

/* - allocates memory for class and DISCLASS event descriptors
   - initializes elements of dc_global data structure
   - sets seed for random generator

   Called by: user program
   Calls    : dc_error
*/

{ 
  int    i;

  strcpy(dc->err_file, err_file);   /* prepare error-handling */
  dc->err_proc = err_proc;
  dc->err_ptr  = err_ptr;
 
  /* allocate memory for class descriptors, one extra as upper limit */
  dc->first = dc->cbeg = calloc(num_classes + 1, sizeof(dc_class));  
  if ( !dc->cbeg )
    dc_error(1, "dfvffvep", dc->err_file, 
	     "Initialization error: unable to allocate memory for %d",
	     num_classes+1, " class descriptors.\n",
	     "(%ld byte required\n)", (num_classes+1)*sizeof(dc_class),
	     dc->err_proc, dc->err_ptr);
  dc->cend = dc->cbeg + num_classes;  /* set ptr to upper limit */
  
  /* allocate memory for event descriptors, no delimiter needed */
  dc->cbeg->bot = calloc(num_deds, sizeof(dc_event));
  if ( !dc->cbeg->bot )
     { 
       free(dc->cbeg); /* if not possible, free all memory allocated */
       dc_error(1, "dfvffvep", dc->err_file, 
		"Initialization error: unable to allocate memory for %d",
		num_deds, " event descriptors.\n",
		"(%ld byte required\n)", num_deds*sizeof(dc_event),
		dc->err_proc, dc->err_ptr);
     }

  /* set pointers from class descriptors to event array, all classes are
     empty, the same amount of memory is reserved for each class, build
     linked list of class descriptors; no mem. for dummy class dc->cend;
     enter class' individual event rates */
  for ( i=0 ; i < num_classes ; i++ )
      { dc->cbeg[i].top = dc->cbeg[i].bot =       /* bot == top means empty  */
 	  dc->cbeg->bot+(num_deds/num_classes)*i; /* each class same portion */
	dc->cbeg[i].next  = &(dc->cbeg[i+1]);      
	dc->cbeg[i].r_ind = class_r_ind[i];
      }

  /* set up empty, limiting class at upper end of class descriptor array */
  dc->cbeg[num_classes].bot =
  dc->cbeg[num_classes].top = dc->cbeg->bot+num_deds;

  /* store pointer to "linking" function into dc */
  if ( event_moved )  
    dc->event_moved = event_moved;
  else 
    dc_error(1, "dfep", dc->err_file, 
	     "Initialization error: No linking function present.",
	     dc->err_proc, dc->err_ptr);
  
  dc->max_events  = num_deds;   /* initialize remaining elements of dc */
  dc->num_classes = num_classes;
  dc->nullclass   = nullclass;
  dc->number_of_reorgs = 0;
  dc->r = 0.0;
  
  seedrand(rand_seed);         /* seed random generator      */
  
}  /* end -- dc_init */

/*--------------------------------------------------------------------------*/

dc_event *dc_move(dc_global *dc, dc_event *ded, void *ued, int new_ci)

/*   - move the event described by ded/ued to class new_ci
     - update ded->ci, class rates of old and new class and
       total rate
     - reorganize memory if necessary

   Called by: user program

   Calls:     dc_delete, dc_store
*/

{ 
  if ( ded )                                /* the event has a ded already  */
    { 
      if ( new_ci == ded->ci )              /* event remains in old class   */
	return ded;                         /* unchanged pointer to ded     */
      else 
	{ 
	  dc_delete(dc, ded);               /* remove event from old class  */
	  ded = dc_store(dc, ued, new_ci);  /* put it in new class, get ded */
	  return ded;                       /* pointer to new ded           */
	}
    }
  else                                  /* the event needs to get a new ded */
    return dc_store(dc, ued, new_ci);
}    /* end -- dc_move */

/*--------------------------------------------------------------------------*/

dc_event  *dc_rand(dc_global *dc)

/* - select one event at random

   Called by: user-program
   Calls    : dc_error
*/

{ 
  double    r;           /* limit for linear selection of class          */
  dc_class  *rand_class, /* ptr to descriptor of class of event selected */
            **preds_next,/* used in sorting linked list                  */
            *old_cd; 
  size_t    class_size;  /* number of events in class selected           */ 
  int       ci;          /* class index                                  */

  /* draw random selection limit, uniformly distributed in 0 <= r < dc_r */
  r = dc->r * doublerand(); 

  rand_class = dc->first;     /* start search at top of list */
  preds_next = &(dc->first);  
  while( rand_class != dc->cend )  /* empty class at end of array AND list */
    { 
      if ( rand_class->next->r > rand_class->r) /* next has higher rate      */
	{                                       /* exchange in list          */
	  old_cd        = rand_class;           /* save ptr to current       */
	  rand_class    = rand_class->next;     /* make next current         */
	  old_cd->next  = rand_class->next; /* update link next->next's next */
	  rand_class->next = old_cd;        /* update link current->next     */
	  *preds_next   = rand_class;  /* update link current's predecessor ->
					current, preds_next points to pred.'s
					next element or dc->first           */
	} /* end -- if ( rand_class-> ... */
      
      r -= rand_class->r;  /* count down r               */
      if ( r <= 0 )        /* selection limit is reached */
	break; /* from while ( rand_class != dc->cend ) */

      /* current class becomes predecessor, current->next becomes current   */
      preds_next = &(rand_class->next); /* get ptr to pred's next           */
      rand_class =   rand_class->next;  /* step to next class in list       */
    }  /* end -- while ( rand_class != dc->cend ) */

  ci = (int) (rand_class - dc->cbeg);    /* index of selected class */

  /* check for class selection errors */
  if ( ci == dc->num_classes )
    dc_error(1, "dffvfvep", dc->err_file,
	     "Error choosing a class by linear selection:\n",
	     "Remainder of random number r = %10.4e\n", r,
	     "index of class selected : %4.0f,\n", (double) ci,
	     dc->err_proc, dc->err_ptr);

  class_size = rand_class->top - rand_class->bot; /*->top is NOT a valid ded */
  if ( class_size <= 0 )
    dc_error(1, "dfep", dc->err_file,
	     "Error in dc_rand: selected class is empty.\n",
	     dc->err_proc, dc->err_ptr);

  /* draw event from class */
  return (rand_class->bot + longrand(class_size));

} /* end -- dc_rand */

/*--------------------------------------------------------------------------*/

dc_event  *dc_store(dc_global *dc, void *ued, int new_ci)

/*   - place the event described by ued in class new_ci
     - store event's, increase class rate, recompute total rate
     - set link ded->ued
     - reorganize memory if necessary

   Called by: user program, dc_move
   Calls    : dc_reorg, dc_error
*/

{
  dc_event *ded;           /* pointer to ded assigned to event */
  dc_class *new_class;

  if ( new_ci == dc->nullclass )
    return NULL;
  
  if ( !dc || !ued )    /* error - dc, ued is not valid */
    dc_error(1, "dfep", dc->err_file,
	     "Error: null pointer for dc or ued passed to dc_store.\n",
	     dc->err_proc, dc->err_ptr);
 
  new_class = dc->cbeg + new_ci;
  ded = new_class->top;            /* top is the first "free" ded */

  /* check for collision with next class */
  if ( ded == (new_class+1)->bot ) 
    { dc_reorg(dc);                          /* reorganize memory   */
      ded = new_class->top;                  /* now it's okay       */
    }

  new_class->top++;                          /* move top            */
  ded->ued = ued;                            /* create link to ued  */  
  ded->ci  = new_ci;                         /* store event's class */
  new_class->r += new_class->r_ind;          /* update class' rate  */
  dc->r += new_class->r_ind;                 /* update total rate   */
 
  return ded;     /* ... pointer to assigned ded */

} /* end -- dc_store */


/* ========================================================================= */
/*                                                                           */
/*          Private section: users should keep their HANDS OFF !!!           */
/*                                                                           */
/* ========================================================================= */

/*   - remove event from its class, reduce class and total rates
       accordingly

   Called by: dc_move
   Calls    : dc->event_moved
*/

void dc_delete(dc_global *dc, dc_event *ded)
{ 
  dc_class  *events_class;
  int       ci;
  
  events_class     = dc->cbeg + ded->ci; 
  events_class->r -= events_class->r_ind;    /* reduce class and */
  dc->r           -= events_class->r_ind;    /* total rate       */

  /* remove ded from class' array of deds */
  events_class->top--; /* ->top, i.e. the delimiting, invalid ded just beyond
			  the last valid ded in the array, goes down by one 
			  as array is kept contiguous */  
  if ( ded != events_class->top )   /* if ded is not at the top of the array, 
				       we got a gap inside */
     { *ded = *(events_class->top);  /* fill the gap by the ded at top */
       dc->event_moved(ded); /* update the link ued->ded for the event moved */
     }

  if ( events_class->top == events_class->bot )
    {                          /* set rate to 0, recompute total rate */
      events_class->r = 0.0;
      dc->r=0.0;
      for ( events_class = dc->cbeg ; events_class<dc->cend ; events_class++ )
	dc->r += events_class->r;
    }
 
}  /* end -- dc_delete */

/*--------------------------------------------------------------------------*/

void dc_error(int error, char *err_fmt, ... )

/* The function dc_error will print a general error/warning message to the
   file given by the user (default is stdout), stating an error in DISCLASS
   and the time; a specific message may be added providing an arbitrary
   number of format strings as well as zero or one double value per format
   string.  Afterwards, an error handler provided by the program using
   DISCLASS may be called via a function-pointer; program execution need not
   be returned from this handler but will be terminated upon returning.
   The flag "error" distinguishes between an error message which causes an
   exit(1) statment at the end of dc_error and a warning which gives the
   message and returns control to the calling routine.
*/

{ 
  va_list arg_ptr;             /* points to next argument in list */
  FILE    *err_file;           
  char    *filename, *format;
  double  value;
  void    (*err_proc)(),
          *err_ptr;
  time_t  err_time;
  int     i=0;                 /* index to format string err_fmt */
  
  va_start(arg_ptr, err_fmt);    /* initialize arg_ptr */
  time(&err_time);  

  /* open file for error message */
  if ( err_fmt[i] == 'd' )
    {                            /* user-defined file */
      i++;
      filename = va_arg(arg_ptr, char *);
      err_file = fopen(filename, "a");
      if ( !err_file )           /* user's file cannot be opened */
	{
	  printf("\n\nERROR: unable to open error log file %s at %s,\n"
		 "       Messages redirected to STDOUT.\n", 
		 filename, ctime(&err_time));
	  err_file = stdout;
	}
    }
  else
    err_file = stdout;           /* stdout is default file */

  /* print general message */
  fprintf(err_file, "\n\n---------------------------------------------------");
  fprintf(err_file, "\n An error occured in the DISCLASS routines");
  fprintf(err_file, "\n at %s\n", ctime(&err_time));

  /* print error-specific message */
  while ( err_fmt[i] == 'f' )     /* process one block [f[v]] */
    { i++;
      format = va_arg(arg_ptr, char *);       /* format string for fprintf */
      if ( err_fmt[i] == 'v' )
	{                         /* process one [v] */
	  i++;
	  value = va_arg(arg_ptr, double);    /* ptr to arg to be printed  */
	  fprintf(err_file, format, value);
	}
      else
	fprintf(err_file, format);      /* print string only */
    }  /* end -- while ( err_fmt ... */

  /* close error file */
  if ( err_file != stdout ) fclose(err_file);

  /* if error==1 call user's own error handler if it exists */
  if ( (err_fmt[i++]=='e') && error )
    { err_proc = va_arg(arg_ptr, void *);    /* get pointer to function     */
      if ( err_fmt[i] == 'p' )
	{                                    /* argument given for err_proc */
	  err_ptr = va_arg(arg_ptr, void *); /* ptr to arg for error proc   */
	  va_end(arg_ptr);                   /* close arg list properly     */
	  if ( err_ptr )          /* call err_proc with arg only if != NULL */
	        err_proc(err_ptr);  
	  else  err_proc();
	}
      else
	{ va_end(arg_ptr);    /* close argument list properly */
	  err_proc();         /* proc without args            */
	}
    }  /* if -- err_fmt[i++] ... */

  /* for error exit program with status 1 indicating proper error handling */
  if ( error ) exit(1);
  
}  /* end -- dc_error */

/*--------------------------------------------------------------------------*/

void dc_reorg(dc_global *dc)
{ size_t    class_size,   /* no. of assigned deds in class        */
            new_size,     /* no. of deds to be allocated to class */
            tot_needed;   /* total number of deds needed          */
  dc_event  *ded;
  dc_class  *cls;
  int       ci;           /* class index                   */

  /* 0. recalculate rates */
  dc->r = 0.0;
  for ( cls = dc->cbeg ; cls < dc->cend ; cls++ )
    {
      cls->r  = (double)( cls->top - cls->bot ) * cls->r_ind;
      dc->r  += cls->r;  
    }

  /* 1. move assigned deds downward, calculate no. of deds needed */
  tot_needed = dc->cbeg->top - dc->cbeg->bot; /* needed for lowest class */
  for ( ci=1 ; ci<dc->num_classes ; ci++ )  
    { class_size = dc->cbeg[ci].top - dc->cbeg[ci].bot; 
      tot_needed += class_size;
      /* move assigned deds of class ci down using memove(to, from, bytes) */
      memmove(dc->cbeg[ci-1].top, dc->cbeg[ci].bot,
	      class_size * sizeof(dc_event));       
      /* update pointers from class descriptors to class' array of deds 
         making use of the contiguous packing of assigned deds          */
      dc->cbeg[ci].bot = dc->cbeg[ci-1].top;
      dc->cbeg[ci].top = dc->cbeg[ci].bot + class_size;
    }  /* end -- for ( ci = 1; ... */
  
  /* add a little extra and reserve 2 deds for classes currently empty */
  tot_needed += (dc->num_classes*2); 

  if ( tot_needed > dc->max_events )    /* out of memory !!! */
    dc_error(1, "dffep", dc->err_file,
	     "Reorganization error: run out of memory ",
	     "while reorganizing event descriptors.\n", 
	     dc->err_proc, dc->err_ptr);
  
  /* 2. move deds up to the new locations class by class, starting from top */
  for ( ci=dc->num_classes-1 ; ci ; ci-- )
    { class_size = dc->cbeg[ci].top - dc->cbeg[ci].bot; 
      new_size   = (size_t) floor((class_size+2.0) * 
				  (dc->max_events / (double) tot_needed));

      /* bottom of class ci's ded-array is reached from bottom of class ci+1
	 (has been moved already) moving down by new_size spaces for deds,
	 top of ci points to first free ded in class' array */
      dc->cbeg[ci].bot = dc->cbeg[ci+1].bot - new_size;
      dc->cbeg[ci].top = dc->cbeg[ci].bot + class_size; 

      /* finally move deds to the right place */
      memmove(dc->cbeg[ci].bot, dc->cbeg[ci-1].top,
	      class_size * sizeof(dc_event));
    }  /* end -- for ( ci = dc-> ... */
  
  /* check if deds have been moved too far down due to round-off errors */
  if ( dc->cbeg[1].bot < dc->cbeg[0].top )
    dc_error(1, "dffep", dc->err_file, 
	     "Reorganization error: re-allocation of ",
	     "event descriptors failed.\n",
	     dc->err_proc, dc->err_ptr);
  
  /* 3. update links ued->ded */
  for ( ci=0 ; ci<dc->num_classes ; ci++ )
      for ( ded=dc->cbeg[ci].bot ; ded<dc->cbeg[ci].top ; ded++ )
          dc->event_moved(ded);
    
  dc->number_of_reorgs++;  /* maintain counter for performance control */
  return;

}  /* end -- dc_reorg */



