/*
   
   THE DISCLASS PACKAGE

   User interface, type declarations and function headers for the 
   discrete classes code.

   Version                    : 1.2
   Last modification          : October 24, 2002

   Documentation and Reference: see MarkovClasses.pdf

*/

/*

  Copyright (C) 1996, 2002 Dietmar Wendt,
                Hans Ekkehard Plesser <hans.plesser@nlh.no>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/* ========================================================================= */
/*                                                                           */
/*                                  Contents                                 */
/*                                                                           */
/* ========================================================================= */

/* This file consists of four parts:

   - Includes and Definition of Constants:
       contains all #include directives required by dc.*
       and defines all preprocessor constants

   - Public type definitions:
       dc_event
       dc_class
       dc_global
     which the user needs to define corresponding variables

   - Public function declarations:
       dc_clear
       dc_init
       dc_move
       dc_rand
       dc_store
     which constitute the user interface of the logclass package

   - Private function declarations [moved to dc.1.2.c]:
       dc_delete
       dc_error
       dc_reorg
      which are internal to logclass and should not be called by the user

*/


/* ========================================================================= */
/*                                                                           */
/*                 Includes and definition of constants                      */
/*                                                                           */
/* ========================================================================= */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <errno.h>
#include <limits.h>   /* compiling with ULTRIX gcc results in warning that   */
#include <float.h>    /* FLT_MAX has been redefined by float.h, but this     */
                      /* does not seem to cause problems                     */

/* modify the following four lines to use your own random generator          */
#include <rand55.h>   /* random generator by Fricke/Knuth                    */
#define seedrand(seed) (init_rand55(seed))   /* seeding command              */
#define longrand(max)  (lrand55(max))        /* long rnd, 0 <= ... < max     */
#define doublerand()   (drand55())           /* double rnd, 0<= ... < 1      */

#ifndef NAME_MAX
#define NAME_MAX 200
#endif


/* ========================================================================= */
/*                                                                           */
/*                        Public type definitions                            */
/*                                                                           */
/* ========================================================================= */

typedef struct DC_EVENT{ void  *ued;  /* pointer to user's event descriptor  */
			 int   ci;    /* index of the event's class          */
		       } dc_event;

/* Type of the DISCLASS event descriptors (deds), representing individual
   events and being the elementary data of the DISCLASS package.

   Pointer to user event descriptor (ued) is void *, so it may be cast to 
   any pointer type and each user may choose her/his own type of ued.

   The user-program needs to read element ued; it may read element ci, but it
   MUST NOT assign to any element; the entries are maintained by the
   DISCLASS functions, so please KEEP YOUR HANDS OFF !!!
*/

/*--------------------------------------------------------------------------*/

typedef struct DC_CLASS{ dc_event         *bot, *top;
			 double           r, r_ind;
			 struct DC_CLASS  *next;
                       } dc_class;

/* Type of class descriptors, each describing one class.

   A discrete class D[z] is the set of all events e with rates

              w_e == w(D[z]) == DC_CLASS.r_ind

   Elements:
     - bot   : first element of class' ded array 
     - top   : pointer to the first ded "behind" this array, i.e. top does 
               not point to a valid ded, but serves as upper limit
     - r     : total rate of the class
     - r_ind : rate of individual event in class
     - next  : links each class to its successor in a list, which is 
               (imperfectly) sorted by class rate (descending order)

   Direct contact to these class descriptors should not be necessary, 
   so please KEEP YOUR HANDS OFF !!!
*/

/*--------------------------------------------------------------------------*/

typedef struct DC_GLOBAL { dc_class           *cbeg, *cend, *first;
			   int                num_classes, nullclass;
			   size_t             max_events;
		           double             r;
		           long int           number_of_reorgs;
		           void               (*event_moved)(dc_event *);
		           char               err_file[NAME_MAX];
			   void               (*err_proc)();
			   void               *err_ptr;
			  } dc_global;

/* The global data structure of DISCLASS containing all information required
   by DISCLASS' functions to operate.  

   The user-program has to allocate memory for and provide a pointer to this
   structure, e.g. defining

     dc_global global_data, *global_data_ptr = &global_data;

   this pointer is passed to all DISCLASS functions as first argument.

   All entries are set by function dc_init and maintained by DISCLASS
   throughout program execution.  They are

   cbeg   : pointer to descriptor of the class with lowest rate, i.e.
            "beginning" of the array of class descriptors
   cend   : pointer to descriptor of an empty class, marking the upper
            limit of the array, i.e. valid pointers to classes are given
	    by
	       cbeg[n] < cend, 0 <= n ( < num_classes )
   first  : pointer to first element in linked list of class descriptors,
            i.e. class where dc_rand starts search
   num_classes: number of class descriptors allocated, i.e. number of elements
                in array beginning at cbeg EXCLUDING dummy class cend.
   nullclass  : null element of set of classes (for impossible events)
   max_events : number of DISCLASS event descriptors (deds) allocated, i.e.
                maximum number of events that can be handled
   r          : total rate of the system, 1/r is mean of the
                first passage time distribution of the Markoff-events
   number_of_reorgs: counts how often memory has been reorganized since
                     initialization 
   *event_moved    : pointer to function defined in user-program to update
                     the link ued->ded
   err_file        : name of file for error messages
   *err_proc       : pointer to function defined in user-program to "wrap
                     things up" before exiting due to error in DISCLASS
   *err_ptr        : in case of error, err_proc is called with err_ptr as
                     argument, so the user can pass a reference to his/her
                     data to the "wrap up routine"; if you have no such
                     pointer, pass NULL.

   The element r is used by the user program to determine the time step 
   according to Gillespie.
   The element number_of_reorgs may be read to monitor "memory performance".

   Any other contact to the global data should not be necessary, 
   so please KEEP YOUR HANDS OFF !!!
*/


/* ========================================================================= */
/*                                                                           */
/*                      Public function declarations                         */
/*                                                                           */
/* ========================================================================= */

void dc_clear(dc_global *dc);

/* Task: 
     - returns memory allocated by dc_init() to operating system

   Arguments: 
     - dc : pointer to global data structure
     
   Called by:
     user-program

   Remarks: 
     - Should be called at the end of simulation
*/

/*--------------------------------------------------------------------------*/

void dc_init(dc_global *dc, size_t num_deds, int num_classes,
	     double *class_r_ind, int nullclass, long int rand_seed,
             void (*event_moved)(), char err_file[], void (*err_proc),
	     void *err_ptr);

/* Task: 
     - allocates memory for class and DISCLASS event descriptors
     - initializes elements of dc_global data structure
     - sets seed for random generator

   Arguments: 
     - dc         : pointer to global data structure
     - num_deds   : number of DISCLASS event descriptors to allocate
     - num_classes: number of classes, excluding null class
     - class_r_ind: array of rates of individual events in classes
     - nullclass  : index of null class for impossible events
     - rand_seed  : random seed
     - event_moved: (pointer to) function updating link ued->ded
     - err_file   : string with name of file for error messages
     - err_proc   : (pointer to) function "wrapping up" after an error
     - err_ptr    : argument to be passed to err_proc 

   Exits with error message if
     - there was not enough memory available
     - event_moved was NULL

   Called by:
     user-program

   Calls:
     dc_error

   Remark: 
     MUST be called before any other DISCLASS routine !
*/

/*--------------------------------------------------------------------------*/

dc_event *dc_move(dc_global *dc, dc_event *ded, void *ued, int new_ci);

/* Task:
     - move the event described by ded/ued to class new_ci
     - update ded->ci, class rates of old and new class and
       total rate
     - reorganize memory if necessary

   Arguments:
     - dc           : pointer to global data structure
     - ded, ued     : pointers to event's ded/ued, respectively
     - new_ci       : NEW class of the event

   Returns:
     - NEW pointer to the event's ded, NULL if new_ci == nullclass

   Exits with error message, if
     - new_ci is out of range of classes
     - dc or ued are NULL
     - all deds allocated have been assigned already
     - memory reorganization failed

   Called by:
     user program

   Calls:
     dc_delete, dc_store

   Remarks:
     - is provided to move events after event execution
     - does NOT update link ued -> ded, since it is faster if the user program
       does it, e.g. by
         ued->ded_ptr = dc_move(dc, ued->ded_ptr, ued, new_ci);
*/

/*--------------------------------------------------------------------------*/

dc_event *dc_rand(dc_global *dc);

/* Task:
     - select one event at random

   Arguments:
     - dc         : pointer to the global data structure

   Returns:
     - dc_rand    : pointer to the ded of the event selected

   Exits with error message and error code 1, if
     - a non-existing class was selected
     - the class selected does not contain any events
     - no class was selected at all

   Called by:
     user-program

   Calls:
     dc_error

   Remarks:
     - as a side effect, dc_rand exchanges adjacent elements of the
       linked list of class descriptors if they are no longer in descen-
       ding order of class rate, so the list remains partially sorted
*/

/*--------------------------------------------------------------------------*/

dc_event  *dc_store(dc_global *dc, void *ued, int new_ci);

/* Task:
     - place the event described by ued in class new_ci
     - store event's, increase class rate, recompute total rate
     - set link ded->ued
     - reorganize memory if necessary

   Arguments:
     - dc          : pointer to the global data structure
     - ued         : pointer to event's ued
     - new_ci      : index of destination class

   Returns:
     - dc_store : pointer to the events ded

   Exits with error message, if
     - dc or ued are NULL
     - all deds allocated have been assigned already
     - memory reorganization failed

   Called by:
     user program, dc_move

   Calls:
     dc_reorg, dc_error

   Remarks:
     - does NOT update link ued->ded
     - if new_ci == nullclass, the NULL pointer is returned
     - should be called by user program ONLY when initializing the data
       structures, i.e. loading the initial configuration
*/


