/*
   
   THE LOGCLASS PACKAGE

   Function definitions of the dual-logarithmic-classes code.

   Idea and first coding      : Thomas Fricke
   Alpha, beta, gamma tests   : Dietmar Wendt
   Polishing & Documentation  : Hans Ekkehard Plesser, D. Wendt
   Maintainer                 : Hans Ekkehard Plesser <hans.plesser@nlh.no>

   Version                    : 1.3
   Last modification          : October 24, 2002

   Header file                : dlc.1.3.h
   Documentation & References : see logclass.pdf

*/

/*

  Copyright (C) 1995, 2002 Thomas Fricke, Dietmar Wendt,
                Hans Ekkehard Plesser <hans.plesser@nlh.no>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/* ========================================================================= */
/*                                                                           */
/*                                  Contents                                 */
/*                                                                           */
/* ========================================================================= */

/* This file consists of two parts:

   - Public function definitions:
       dlc_init
       dlc_clear
       dlc_enter
       dlc_rand
       dlc_safechange
       dlc_resum
     which constitute the user interface of the logclass package

   - Private function definitions:
       dlc_delete
       dlc_do_absolutely_nothing
       dlc_error
       dlc_getclass
       dlc_reorg
       dlc_store
       dlc_knownchange
       dlc_unknownchange
      which are internal to logclass and should not be called by the user
*/

/* ========================================================================= */
/*                                                                           */
/*                              Revision History                             */
/*                                                                           */
/* ========================================================================= */

/*
  1.3  Hans Ekkehard Plesser 2002-10-24

  - order of checks in dlc_rand changed
  - dlc_resum called if round-off problems detected or class becomes empty
  - private function declarations moved from dlc.*.h to dlc.*.c
  
*/


/* ========================================================================= */
/*                                                                           */
/*                                  Includes                                 */
/*                                                                           */
/* ========================================================================= */

#include "dlc.h"        /* contains all further includes, definition of
		           preprocessor commands, type definitions and
		           public function declarations                      */
#include <rand55.h>     /* Knuth's Lagged Fibonacci generator, implementation
                           by T. Fricke                                      */

/* ========================================================================= */
/*                                                                           */
/*          Private function declaratoins: for internal use only             */
/*                                                                           */
/* ========================================================================= */

/* moved to dlc.1.3.c, so users do not get tempted to call these functions
   from their code
*/

/*--------------------------------------------------------------------------*/

void dlc_delete(dlc_global *dlc, dlc_class *events_class, dlc_event *led);

/* Task:
     - remove event from its class, reduce class and total reactivity
       accordingly

   Arguments:
     - dlc         : pointer to the global data structure
     - events_class: ptr to descriptor of current class of event to be removed
     - led         : ptr to led of event

   Exits with error message, if
     - round-off errors are detected

   Called by:
     dlc_knownchange

   Calls:
     dlc->event_moved, dlc_error

   Remarks:
     - dlc_delete does not verify that the event led is indeed a member
       of class events_class; if this is not the case, unpredictable
       things may happen!!!
*/



/*--------------------------------------------------------------------------*/
void dlc_do_absolutely_nothing(dlc_event *led);

/* Purpose:
   A dummy routine assigned to dlc->event_moved if dlc_init is called with
   argument event_moved == NULL.  This keeps everything well-defined if the
   user-program does not provide a function for updating the link ued->led,
   but as soon as the first led is moved in memory, the ueds cannot "find"
   their leds anymore and the algorithm will havoc.  Thus, it makes sense only
   if all events have constant reactivities and do not move.  In that case,
   on the other hand, you should not use the LOGCLASS package at all but
   apply von-Neumann-rejection to a pre-sorted array.

   Arguments:
     - led : pointer to led of an event that has moved

   Called by (using pointer dlc->event_moved):
     dlc_delete, dlc_reorg
*/



/*--------------------------------------------------------------------------*/
void dlc_error(int error, char *err_fmt, ... );

/* Task:
     - issues an error/warning message and terminates execution by error
     - if a pointer to a function is passed as argument, this function
       is called before calling exit() if error is 1

   Arguments:
     - error : Flag which distinguish between an error and warning.
               An error causes an exit-statment at the end of dlc_error
     - err_fmt == "[d][f[v]][f[v]]...[f[v]][e[p]]"
         format string indicating the content of the variable argument list:
	   + one lowercase character per argument, no spaces
	   + [] indicate optional entries, ... repetition
	   + entries must be in the order given above
	     d: name of file to which error message is to be written
	     f: format string to be passed to fprintf, may contain
	        zero or one conversion characters for DOUBLE 
		values, i.e. %___e, %___f or %___g
	     v: DOUBLE value to be used as argument by fprintf
	     e: pointer to user's own error handling function:
	        must return void, may take zero or one ptr as argument
	     p: pointer of type void* to be passed to users error function
                if !=NULL

     - ...  : arguments as specified in err_fmt

   Returns:
     - if error==0 gives its warning message and returns
     - if error==1 does NOT return control, but calls exit(1) !

   Called by:
     dlc_delete, dlc_getclass, dlc_init, dlc_rand, dlc_reorg, dlc_store

   Calls:
     function of type void() if passed as argument

   Remarks:
     - the function passed as argument does not have to return control to
       dlc_error; in this case, though, the program may terminate with 
       status != 1
     - format string and argument list must match; there is
       no error-checking provided by dlc_error.
*/



/*--------------------------------------------------------------------------*/
dlc_class *dlc_getclass(dlc_global *dlc, double r);

/* Task:
     - determines the class that comprises the reactivity r

   Arguments:
     - dlc : pointer to the global data structure
     - r   : reactivity

   Returns:
     - dlc_getclass: pointer to the descriptor of the corresponding class
                     NULL if r<= minimum of lowest class

   Exits with error message, if
     - frexp() sets errno
     - r < 0
     - reactivity is out of range of classes

   Called by:
     dlc_enter, dlc_knownchange, dlc_unknownchange

   Calls:
     dlc_error

   Remarks:
     - if r < 2^(dlc->min_class-1), the NULL pointer is returned
*/



/*--------------------------------------------------------------------------*/
void dlc_reorg(dlc_global *dlc);

/* Task:
     - reorganizes memory when a class gets overcrowded
     - after leds have been moved in memory, dlc->event_moved is used
       to update the ued->led link
     - recalculates class and total reactivities to reduce round-off errors

   Arguments:
     - dlc : pointer to the global data structure

   Exits with error message, if
     - all leds allocated have been assigned already
     - memory reorganization failed otherwise

   Called by:
     dlc_store

   Calls:
     dlc->event_moved, dlc_error
*/



/*--------------------------------------------------------------------------*/
dlc_event  *dlc_store(dlc_global *dlc, dlc_class *events_class, void *ued, 
		      double r);

/* Task:
     - place the event described by ued in the class given by events_class
       according to its NEW reactivity r
     - store event's, increase class, total reactivity by r
     - set link led->ued
     - reorganize memory if necessary

   Arguments:
     - dlc         : pointer to the global data structure
     - events_class: pointer to descriptor of class in which to place event
     - ued         : pointer to event's ued
     - r           : NEW reactivity of the event

   Returns:
     - dlc_store : pointer to the events led

   Exits with error message, if
     - dlc or ued are NULL
     - all leds allocated have been assigned already
     - memory reorganization failed

   Called by:
     dlc_enter, dlc_knownchange

   Calls:
     dlc_reorg, dlc_error

   Remarks:
     - dlc_store does not verify if the event belongs to the class given;
       if it does not, this will lead to wrong results and might cause
       an error message by dlc_rand after some time
     - does NOT update link ued->led
     - if events_class == NULL, the NULL pointer is returned
*/


/*--------------------------------------------------------------------------*/
dlc_event *dlc_knownchange(dlc_global *dlc, dlc_class *events_class,
			   dlc_event *led, void *ued, double r);

/* Task:
     - move the event described by led/ued whose reactivity r has changed
       to the appropriate class if its current class is known;
     - update event's reactivity entry in led, class reactivities of all
       classes involved and total reactivity
     - reorganize memory if necessary

   Arguments:
     - dlc          : pointer to global data structure
     - events_class : pointer to descriptor of event's current class
     - led, ued     : pointers to event's led/ued, respectively
     - r            : NEW reactivity of the event

   Returns:
     - NEW pointer to the event's led, NULL if r==0

   Exits with error message, if
     - round-off errors are detected
     - frexp() sets errno
     - r < 0
     - reactivity is out of range of classes
     - dlc or ued are NULL
     - all leds allocated have been assigned already
     - memory reorganization failed

   Called by:
     user-program, dlc_safechange, dlc_unknownchange

   Calls:
     dlc_delete, dlc_getclass, dlc_store

   Remarks:
     - led MUST BE != NULL, otherwise behavior is unpredictable; 
       if led == NULL might occur, use dlc_safechange !
     - is provided to move the event selected by dlc_rand after execution
     - argument events_class is provided by dlc_knownchange as re_class
     - does NOT update link ued -> led, since it is faster if the user program
       does it, e.g. by
         ued->led_ptr = dlc_knownchange(dlc, events_class, led, ued, r);
*/


/*--------------------------------------------------------------------------*/
dlc_event *dlc_unknownchange(dlc_global *dlc, dlc_event *led, void *ued, 
			     double r);

/* Task:
     - move the event described by led/ued whose NEW reactivity is r to 
       the appropriate class
     - update event's reactivity entry in led, class reactivities of all
       classes involved and total reactivity

   Arguments:
     - dlc          : pointer to global data structure
     - led, ued     : pointers to events led/ued, respectively
     - r            : NEW reactivity of the event

   Returns:
     - NEW pointer to the event's led, NULL if r==0

   Exits with error message, if
     - round-off errors are detected
     - frexp() sets errno
     - r < 0
     - reactivity is out of range of classes
     - dlc or ued are NULL
     - all leds allocated have been assigned already
     - memory reorganization failed

   Called by:
     not used in this version

   Calls:
     dlc_getclass, dlc_knownchange

   Remarks:
     - The same as dlc_knownchange, except the current class of the event 
       needs not to be specified.
     - led MUST BE != NULL, otherwise behavior is unpredictable; 
       if led == NULL might occur, use dlc_safechange !
     - does NOT update link ued -> led, since it is faster if the user program
       does it, e.g. by
         ued->led_ptr = dlc_unknownchange(dlc, led, ued, r);
*/


/* ========================================================================= */
/*                                                                           */
/*                              Public functions                             */
/*                                                                           */
/* ========================================================================= */

void dlc_init(dlc_global *dlc, size_t num_leds, double min_reac, 
	      double max_reac, long int rand_seed, void (*event_moved)(), 
	      char err_file[], void (*err_proc)(), void *err_ptr)

/* - allocates memory for class and LOGCLASS event descriptors
   - initializes elements of dlc_global data structure
   - sets seed for random generator

   Called by: user-program
   Calls    : dlc_error
*/

{ 
  int    i, classlow, classhigh, classn;
  double top_reac;

  strcpy(dlc->err_file, err_file);   /* prepare error-handling */
  dlc->err_proc = err_proc;
  dlc->err_ptr  = err_ptr;

  if ( min_reac <= 0 )
    dlc_error(1, "dfv", dlc->err_file, "Invalid parameter to dlc_init: "
	      "min_reac > 0 required (is %10.4e)\n", min_reac);

  /* determine range of classes required */
  frexp(min_reac, &classlow);  classlow  -= DLC_EXTRA_BOTTOM;
  frexp(max_reac, &classhigh); classhigh += DLC_EXTRA_TOP;
  classn = classhigh - classlow + 1;       /* number of classes required */

  /* check if range is valid, i.e. small enough to avoid round-off */
  dlc->eps = ldexp(0.5, classlow);  /* lowest reactivity > 0 */
  top_reac = ldexp(1.0, classhigh); /* highest reactivity represented */

  /* check if enough leds are requested */
  if ( num_leds <= 2*classn ) {
    dlc_error(0, "dfff", dlc->err_file, "Initialization warning:\n",
	      "On the average, you have less than two events per class.\n",
	      "You should use v. Neumann rejection to avoid overhead!\n");
    num_leds = 3*classn;
  }

  /* allocate memory for class descriptors, one extra as upper limit */
  dlc->first=dlc->cbeg=calloc(classn + 1, sizeof(dlc_class));
  if ( !dlc->cbeg )
    dlc_error(1, "df", dlc->err_file, "Initialization error: unable to "
	      "allocate memory for class descriptors.\n");
  dlc->cend = dlc->cbeg + classn; /* get pointer to upper limit */
  
  /* allocate memory for event descriptors, no delimiter needed */
  dlc->cbeg->bot = calloc(num_leds, sizeof(dlc_event));
  if ( !dlc->cbeg->bot )
    { free(dlc->cbeg); /* if not possible, free all memory allocated */
      dlc_error(1, "df", dlc->err_file, "Initialization error: unable to "
		"allocate memory for event descriptors.\n");
    }
  
  /* set pointers from class descriptors to event array, all classes are
     empty, the same amount of memory is reserved for each class, build
     linked list of class descriptors; no mem. for dummy class dlc->cend */
  for ( i=0 ; i<classn ; i++ )
    { dlc->cbeg[i].top = dlc->cbeg[i].bot =     /* bot == top means empty  */
	dlc->cbeg->bot+(num_leds/classn)*i;     /* each class same portion */
      dlc->cbeg[i].next = dlc->cbeg + (i + 1);
      dlc->cbeg[i].eps = dlc->eps;
    }

  /* set up empty, limiting class at upper end of class descriptor array */
  dlc->cbeg[classn].bot = dlc->cbeg[classn].top = dlc->cbeg->bot+num_leds;
  dlc->cbeg[classn].next = NULL;        

  /* enter "linking" function into dlc, "do_nothing" is default */
  if ( event_moved )
    dlc->event_moved=event_moved;
  else
    dlc->event_moved=dlc_do_absolutely_nothing;
  
  dlc->max_events  = num_leds;   /* initialize remaining elements of dlc */
  dlc->min_class   = classlow;
  dlc->num_classes = classn;
  dlc->number_of_reorgs = 0;
  dlc->r=0; 
  
  init_rand55(rand_seed);     /* seed random generator      */
  
}  /* end -- dlc_init */



/*--------------------------------------------------------------------------*/
void dlc_clear(dlc_global *dlc) /* if dlc not longer needed */

/* - returns memory allocated by dlc_init() to OS
     
   Called by: user-program
*/

{ 
  free(dlc->cbeg->bot);   /* free event descriptors first, then class descs. */
  free(dlc->cbeg);        /* i.e. reverse allocation order                   */
}



/*--------------------------------------------------------------------------*/
dlc_event  *dlc_enter(dlc_global *dlc, void *ued, double r)

/* - assigns a led to the event represented by ued and places it in a
     class according to its reactivity r; sets up link led->ued and
     stores reactivity in led

  Called by: user-program, dlc_safechange
  Calls    : dlc_getclass, dlc_store
*/

{ dlc_class *events_class;
  
  events_class = dlc_getclass(dlc, r);          /* determine event's class */
  return dlc_store(dlc, events_class, ued, r);  /* get a led and return it */
}



/*--------------------------------------------------------------------------*/
dlc_event *dlc_rand(dlc_global *dlc)

/* - select one event at random as described by Fricke&Wendt
   - as a side effect, dlc_rand exchanges adjacent elements of the linked
     list of class descriptors if they are no longer in descending order
     of class reactivity, so the list remains partially sorted

   Called by: user-program
   Calls    : dlc_error
*/

{ double    r;          /* limit for linear selection of class          */
  dlc_event *re;        /* ptr to led of event selected                 */
  dlc_class *re_class,  /* ptr to descriptor of class of event selected */
            **preds_next,  /* used in sorting linked list                  */
            *old_cd; 
  size_t    class_size; /* number of events in class selected           */ 
  int       ci;         /* class index */
  
  /* Select a class by linear selection */

  /* draw random selection limit, uniformly distributed in 0 <= r < dlc_r */
  r = dlc->r * drand55(); 

  /* search linked list of classes reactivites of classes searched from r 
     until r <= 0, sort along the way */
           
  re_class  = dlc->first;     /* start at top of list, not at dlc->cbeg    */
  preds_next= &(dlc->first);  /* if first class in list is exchange, store 
				 ptr to new top's descriptor in dlc->first */
  
  while( re_class != dlc->cend )  /* empty class at end of array AND list */
    { 
      if ( re_class->next->r > re_class->r) /* next class has higher react. */
	{                                   /* exchange descriptors in list */
	  old_cd        = re_class;         /* save ptr to current class    */
	  re_class      = re_class->next;   /* make next class current      */
	  old_cd->next  = re_class->next;   /* update link next->next's next*/
	  re_class->next= old_cd;           /* update link current->next    */
	  *preds_next   = re_class;  /* update link current's predecessor ->
					current, preds_next points to pred.'s
					next element or dlc->first          */
	} /* end -- if ( re_class-> ... */
      
      r -= re_class->r;  /* count down r               */
      if ( r <= 0 )      /* selection limit is reached */
	break; /* from while ( re_class != dlc->cend ) */

      /* current class becomes predecessor, current->next becomes current */
      preds_next = &(re_class->next); /* get ptr to pred's next element   */
      re_class   = re_class->next;    /* step to next class in list       */
    }  /* end -- while ( re_class != dlc->cend ) */
  
  class_size = re_class->top - re_class->bot; /* ->top is NOT a valid led */
  ci         = ( re_class - dlc->cbeg ) + dlc->min_class; 
  /*             ^^^^^^^^^^^^^^^^^^^^     ^^^^^^^^^^^^^^
		      pointers               integer
     index is given by position in array of class descs. plus offset */
  
  /* check for class selection errors */
  if ( re_class == dlc->cend )  /* <=> ci == dlc->num_classes+dlc->min_class */
    dlc_error(1, "dfep", dlc->err_file, 
	      "Error in dlc_rand: no class selected.\n",
	      dlc->err_proc, dlc->err_ptr);
  if ( ci > dlc->num_classes+dlc->min_class )
    dlc_error(1, "dffvfvfvep", dlc->err_file,
	      "Error choosing a class by v. Neumann rejection:\n",
	      " random number r = %10.4e\n", r,
	      " index of class selected : %4.0f,\n", (double) ci,
	      " range is 0 <= ... < %4.0f .\n", (double) dlc->num_classes,
	      dlc->err_proc, dlc->err_ptr);
  if ( class_size <= 0 )
    dlc_error(1, "dfep", dlc->err_file, 
	      "Error in dlc_rand: selected class is empty.\n",
	      dlc->err_proc, dlc->err_ptr);

  /* Select an event from re_class by von Neumann rejection, upper bound
     is maximum reactivity represented by re_class, i.e. 2^ci            */ 
  do 
    { 
      /* draw any event from re_class, uniformly distributed */
      re = re_class->bot+lrand55(class_size); 
    } while ( ldexp(drand55(), ci) > re->r );   /* test rejection criterium */
  /* using ldexp here is faster than calculating x=2^ci before entering the
     loop and multiplying drand55()*x every time; besides, average acceptance
     is 75% */
  
  /* return */
  return re;                  /* pointer to the event's led              */

} /* end -- dlc_rand */



/*--------------------------------------------------------------------------*/
dlc_event *dlc_safechange(dlc_global *dlc, dlc_event *led, void *ued, double r)

/* - move or enter the event described by led/ued whose NEW reactivity is r
     to the appropriate class
   - update event's reactivity, class reactivity, total reactivity
   - reorganize memory if necessary

   Called by: user-program
   Calls    : dlc_enter, dlc_knownchange
*/
     
{ 
  if ( led )      /* the event has a led already */
    return dlc_knownchange(dlc, dlc_getclass(dlc, led->r), led, ued ,r);
  else            /* the event has to be assigned a led */
    return dlc_enter(dlc, ued, r);
}



/*--------------------------------------------------------------------------*/
int  dlc_resum(dlc_global *dlc)

/* - resummarize the reactivities of all not-empty classes
   - set the reactivity of empty classes to 0
   - resummarize dlc->r with the new class-reactivities

   Called by: user-program
*/

{ dlc_class *class;
  dlc_event *led;
  int       r_error=0;
  double    new_class_r;

  dlc->r = 0.0;
                 /* controll all classes for entries */
  for ( class=dlc->cbeg ; class<dlc->cend ; class++ )
      {       /* new calculation of class->r if the class is not empty */
        if ( class->top != class->bot )
	     { new_class_r = 0.0;
               for ( led=class->bot ; led<class->top ; led++ )
                   new_class_r += led->r;
               if ( new_class_r != class->r ) r_error++;
               class->r = new_class_r;
                        /* update dlc->r with new class-reactivity */
               dlc->r += new_class_r;
             }
        else { if ( class->r ) { r_error++;   class->r = 0.0; }
             }
      }
  return r_error;

}



/* ========================================================================= */
/*                                                                           */
/*        Private functions: users should keep their HANDS OFF !!!           */
/*                                                                           */
/* ========================================================================= */



/*--------------------------------------------------------------------------*/
void dlc_delete(dlc_global *dlc, dlc_class *events_class, dlc_event *led)

/* - remove event from its class, reduce class and total reactivity
     accordingly

   Called by: dlc_knownchange
   Calls    : dlc->event_moved, dlc_error
*/

{ double ci;      /* class index */

  if ( !events_class || !led )   /* just to be on the safe side */
    return;
  
  events_class->r -= led->r;  /* reduce class and */
  dlc->r          -= led->r;  /* total reactivity */

  /* remove led from class' array of leds F[] */
  events_class->top--; /* ->top, i.e. the delimiting, invalid led just beyond
			  the last valid led in the array, goes down by one 
			  as array is kept contiguous */  
  if ( led != events_class->top )   /* if led is not at the top of the array, 
				       we got a gap inside */
    { 
      *led = *(events_class->top);  /* fill the gap by the led at top */
      dlc->event_moved(led); /* update the link ued->led for the event moved */
    }

  /* check for accumulated round-off errors */
  if ( events_class->top == events_class->bot )
     {
       if ( events_class->r >= events_class->eps ) 
            { ci = (double) ( events_class - dlc->cbeg );
              dlc_error(0, "dfvfvfvfvep", dlc->err_file,
	  	        "Warning in dlc_delete: class %4.0f is empty but\n",
                        (double) ci,
                        "reactivity = %10.4e > ", events_class->r,
                        "class_eps = %10.4e\n", events_class->eps,
                        "events_class->r_max = %10.4e\n", events_class->r_max,
		        dlc->err_proc, dlc->err_ptr);
            }
       events_class->r = 0;
       events_class->r_max = 0;
       events_class->eps = dlc->eps;
       dlc_resum(dlc);
     }

}  /* end -- dlc_delete */



/*--------------------------------------------------------------------------*/
void dlc_do_absolutely_nothing(dlc_event *led)

{
  /* as the name says ... */
}



/*--------------------------------------------------------------------------*/
void dlc_error(int error, char *err_fmt, ... )

/* The function dlc_error will print a general error/warning message to the
   file given by the user (default is stdout), stating an error in LOGCLASS
   and the time; a specific message may be added providing an arbitrary
   number of format strings as well as zero or one double value per format
   string.  Afterwards, an error handler provided by the program using
   LOGCLASS may be called via a function-pointer; program execution need not
   be returned from this handler but will be terminated upon returning.
   The flag "error" distinguishes between an error message which causes an
   exit-statment at the end of dlc_error and a warning which gives the
   message and continues with the run.
*/

{ va_list arg_ptr;             /* points to next argument in list */
  FILE    *err_file;           
  char    *filename, *format;
  double  value;
  void    (*err_proc)(),
          *err_ptr;
  time_t  err_time;
  int     i=0;                /* index to format string err_fmt */
  

  va_start(arg_ptr, err_fmt);    /* initialize arg_ptr */
  time(&err_time);  

  /* open file for error message */
  if ( err_fmt[i] == 'd' )
    {                            /* user-defined file */
      i++;
      filename = va_arg(arg_ptr, char *);
      err_file = fopen(filename, "a");
      if ( !err_file )           /* user's file cannot be opened */
	{
	  printf("\n\nERROR: unable to open error log file %s at %s,\n"
		 "       Messages redirected to STDOUT.\n", 
		 filename, ctime(&err_time));
	  err_file = stdout;
	}
    }
  else
    err_file = stdout;           /* stdout is default file */

  /* print general message */
  fprintf(err_file, "\n\n---------------------------------------------------");
  fprintf(err_file, "\n An error occured in the dual-log-class routines");
  fprintf(err_file, "\n at %s\n", ctime(&err_time));

  /* print error-specific message */
  while ( err_fmt[i] == 'f' )     /* process one block [f[v]] */
    { i++;
      format = va_arg(arg_ptr, char *);       /* format string for fprintf */
      if ( err_fmt[i] == 'v' )
	{                         /* process one [v] */
	  i++;
	  value = va_arg(arg_ptr, double);    /* ptr to arg to be printed  */
	  fprintf(err_file, format, value);
	}
      else
	fprintf(err_file, format);      /* print string only */
    }  /* end -- while ( err_fmt ... */

  /* close error file */
  if ( err_file != stdout ) fclose(err_file);

  /* if error==1 call user's own error handler if it exists */
  if ( (err_fmt[i++]=='e') && error )
    { err_proc = va_arg(arg_ptr, void *);    /* get pointer to function     */
      if ( err_fmt[i] == 'p' )
	{                                    /* argument given for err_proc */
	  err_ptr = va_arg(arg_ptr, void *); /* ptr to arg for error proc   */
	  va_end(arg_ptr);                   /* close arg list properly     */
	  if ( err_ptr )          /* call err_proc with arg only if != NULL */
	        err_proc(err_ptr);  
	  else  err_proc();
	}
      else
	{ va_end(arg_ptr);    /* close argument list properly */
	  err_proc();         /* proc without args            */
	}
    }  /* if -- err_fmt[i++] ... */

  /* for error exit program with status 1 indicating proper error handling */
  if ( error ) exit(1);
  
}  /* end -- dlc_error */



/*--------------------------------------------------------------------------*/
dlc_class *dlc_getclass(dlc_global *dlc, double r)

/* - determines the class that comprises the reactivity r
   - returns NULL if reactivity is lower than any class's minimum

   Called by: dlc_enter, dlc_knownchange, dlc_unknownchange
   Calls    : dlc_error
*/
{ int    ci;    /* class index */
  
  if ( r < 0 )  /* reactivity must be positive! */
    dlc_error(1, "dffvep", dlc->err_file,"Error determining class of an ",
	      "event:\nreactivity r = %10.4e < 0 .\n", r,
              dlc->err_proc, dlc->err_ptr);
	     
  /* calculate class index */
  errno = 0;       /* reset before use                               */
  frexp(r,&ci);    /* class index is defined by 2^(ci-1) <= r < 2^ci */
  ci -= dlc->min_class;  /* subtract offset to get class-array index */

  /* check for range errors in frexp*/
  if ( errno == ERANGE )     /* r to large for frexp */      
    dlc_error(1, "dfffvep", dlc->err_file,"Error determining class of an ",
	      "event:\n frexp(r, &ci) set errno ERANGE when called with ",
              "r=%10.4e .\n", r, dlc->err_proc, dlc->err_ptr);

  /* check if reactivity is smaller than minimum of lowest class or == 0 */
  if ( ci < 0 || r == 0 )  
    return NULL;   /* and thus is to be removed from LOGCLASS data       */
  
  /* check if reactivity is beyond highest class */
  if ( ci >= dlc->num_classes )
    dlc_error(1, "dffvfvfvfvep", dlc->err_file, 
	      "Error determining class of an event:\n", 		
	      " reactivity r = %10.4e\n", r,
	      " => class index = %4.0f,\n", (double) (ci + dlc->min_class),
	      " range is %4.0f <= ... ", (double) dlc->min_class,
	      "<= %4.0f .\n", (double) (dlc->min_class + dlc->num_classes - 1),
	      dlc->err_proc, dlc->err_ptr);

  /* if ok, return pointer to class descriptor */
  return dlc->cbeg + ci;

} /* end -- dlc_getclass */



/*--------------------------------------------------------------------------*/
void dlc_reorg(dlc_global *dlc)

/* This function is called if all leds allocated for a class have been
   assigned to events and dlc_store requests another led in that class.
   Then, dlc_reorg re-distributes the dlc->max_events leds allocated by
   dlc_init to the classes proportional to their current size.  In the
   end, the order of leds currently assigned to events is not changed,
   just the "free" leds have been shifted around.  All links are main-
   tained properly.
   To do so, dlc_rand employs the following strategy:
     0. Recalculate class and total reactivites to reduce round-off errors.
     1. Move all assigned leds downward in memory, so they occupy a
        contiguous are in memory between dlc->cbeg->bot and 
	dlc->cbeg[dlc->num_classes-1]->top and all "free" leds occupy
	memory from there to dlc->cend.
     2. Move the leds for each class upward to their new location, beginning
        with the highest class.
     3. Update the links ued->led for all events using dlc->event_moved.

   Called by: dlc_store
   Calls    : dlc->event_moved, dlc_error
*/

{ size_t    class_size,   /* no. of assigned leds in class        */
            new_size,     /* no. of leds to be allocated to class */
            tot_needed;   /* total number of leds needed          */
  dlc_event *led;
  dlc_class *cd;
  int       ci;           /* class index                   */

  /* 0. recalculate reactivities */
  for ( cd = dlc->cbeg, dlc->r = 0; cd < dlc->cend; cd++ )
    {
      for ( led = cd->bot, cd->r = 0; led < cd->top; led++ )
	cd->r += led->r;
      dlc->r += cd->r;
    }

  /* 1. move assigned leds downward, calculate no. of leds needed */
  tot_needed = dlc->cbeg->top - dlc->cbeg->bot; /* needed for lowest class */
  for ( ci = 1; ci < dlc->num_classes; ci++)  
    {
      class_size = dlc->cbeg[ci].top - dlc->cbeg[ci].bot; 
      tot_needed += class_size;
      /* move assigned leds of class ci down using memove(to, from, bytes) */
      memmove(dlc->cbeg[ci-1].top, dlc->cbeg[ci].bot,
	      class_size * sizeof(dlc_event));       
      /* update pointers from class descriptors to class' array of leds 
         making use of the contiguous packing of assigned leds          */
      dlc->cbeg[ci].bot = dlc->cbeg[ci-1].top;
      dlc->cbeg[ci].top = dlc->cbeg[ci].bot + class_size;
    }  /* end -- for ( ci = 1; ... */
  
  /* add a little extra and reserve 2 leds for classes currently empty */
  tot_needed += dlc->num_classes * 2; 

  if ( tot_needed > dlc->max_events )    /* out of memory !!! */
    dlc_error(1, "dfep", dlc->err_file, "Reorganization error: run out of "
	      "memory while reorganizing event descriptors.\n", dlc->err_proc,
	      dlc->err_ptr);
  
  /* 2. move leds up to the new locations class by class, starting from top */
  for ( ci = dlc->num_classes - 1; ci ; ci--)
    {
      class_size = dlc->cbeg[ci].top - dlc->cbeg[ci].bot; 
      new_size   = (size_t) floor((class_size+2.0) * 
				  (dlc->max_events / (double) tot_needed));

      /* bottom of class ci's led-array is reached from bottom of class ci+1
	 (has been moved already) moving down by new_size spaces for leds,
	 top of ci points to first free led in class' array */
      dlc->cbeg[ci].bot = dlc->cbeg[ci+1].bot - new_size;
      dlc->cbeg[ci].top = dlc->cbeg[ci].bot + class_size; 

      /* finally move leds to the right place */
      memmove(dlc->cbeg[ci].bot, dlc->cbeg[ci-1].top,
	      class_size * sizeof(dlc_event));
    }  /* end -- for ( ci = dlc-> ... */
  
  /* check if leds have been moved too far down due to round-off errors */
  if ( dlc->cbeg[1].bot < dlc->cbeg[0].top )
    dlc_error(1, "dfep", dlc->err_file, "Reorganization error: re-allocation "
	      "of event descriptors failed.\n", dlc->err_proc, dlc->err_ptr);
  
  /* 3. update links ued->led */
  for ( ci = 1; ci < dlc->num_classes; ci++ )
    for ( led = dlc->cbeg[ci].bot;  led < dlc->cbeg[ci].top; led++ )
      dlc->event_moved(led);
    
  dlc->number_of_reorgs++;  /* maintain counter for performance control */
  return;
}  /* end -- dlc_reorg */



/*--------------------------------------------------------------------------*/
dlc_event  *dlc_store(dlc_global *dlc, dlc_class *events_class, void *ued, 
		      double r)

/* - places the event described by ued in the class given by events_class
     according to its NEW reactivity r
   - stores event's, increases class, total reactivity by r
   - sets link led->ued
   - does NOT update link ued->led
   - reorganize memory if necessary

   Called by: dlc_enter, dlc_knownchange
   Calls    : dlc_reorg, dlc_error
*/

{ dlc_event *led;   /* pointer to led assigned to event */

  if ( !events_class )   /* just to be on the safe side */
    return NULL;
  
  if ( !dlc || !ued )    /* error - dlc, ued is not valid */
    dlc_error(1, "dfep", dlc->err_file, 
	      "Error: null pointer passed to dlc_store.\n", 
	      dlc->err_proc, dlc->err_ptr);
  
  led = events_class->top;   /* top is the first "free" led */
  
  /* check for collision with next class */
  if ( led == (events_class+1)->bot ) 
    { 
      dlc_reorg(dlc);          /* reorganize memory */
      led = events_class->top; /* now it's okay     */
    }
  
  events_class->top++;   /* move top                   */
  led->ued = ued;        /* create link to ued         */  
  led->r   = r;          /* store event's reactivity   */
  events_class->r += r;  /* update classes' reactivity */
  dlc->r          += r;  /* update total reactivity    */
                         /* update maximum reactivity the class ever had  */
                         /* since it was an empty class and calculate the */
                         /* the class-specific eps                        */
  if ( events_class->r > events_class->r_max )
     { events_class->r_max = events_class->r;
       events_class->eps = events_class->r_max*DBL_EPSILON;
       if ( events_class->eps < dlc->eps ) events_class->eps = dlc->eps;
     }

  return led;     /* ... pointer to assigned led */

} /* end -- dlc_store */



/*--------------------------------------------------------------------------*/
dlc_event *dlc_knownchange(dlc_global *dlc, dlc_class *events_class,
			   dlc_event *led, void *ued, double r)

/* - move the event described by led/ued whose reactivity r has changed
     to the appropriate class if its current class is known;
   - update event's reactivity, class reactivity, total reactivity
   - reorganize memory if necessary

   Called by: user-program, dlc_safechange, dlc_unknownchange
   Calls    : dlc_delete, dlc_getclass, dlc_store
*/

{ dlc_class *events_new_class;
  
  /* determine class according to NEW reactivity */
  events_new_class = dlc_getclass(dlc, r);

  if ( events_new_class == events_class )  /* event remains in old class */
    {                                      /* just update reactivities   */
      events_class->r += (r - led->r);     /* class reactivity           */
      dlc->r          += (r - led->r);     /* total     "                */
      led->r           = r;                /* event's reactivity         */

      /* check for accumulated round-off errors and throw error */
      if ( events_class->r < events_class->eps )      
	{ 
	  /* see if we can recover */
	  dlc_resum(dlc);

	  if ( events_class->r < events_class->eps )      
	    {
	      double ci = (double) ( events_class - dlc->cbeg );
	      dlc_error(1, "dfvfvfvfvep", dlc->err_file,
			"Error in dlc_knownchange: non-empty class %4.0f has \n",
			(double) ci,
			"vanishing reactivity = %10.4e < ", events_class->r,
			"class_eps = %10.4e\n", events_class->eps,
			"events_class->r_max = %10.4e\n", events_class->r_max,
			dlc->err_proc, dlc->err_ptr);
	    }
	}

      if ( dlc->r < dlc->eps )
	{
	  dlc_resum(dlc);   /* let's hope we get errors sorted out */
	  if ( dlc->r < dlc->eps )
	    { 
	      double ci = (double) ( events_class - dlc->cbeg );
	      dlc_error(1, "dfvfvfvep", dlc->err_file,
		 "Error in dlc_knownchange:\n Vanishing total reactivity = %10.4e < ", 
			dlc->r,
			"eps = %10.4e\n", dlc->eps,
                        "in spite of non-empty cell ",
			(double) ci,
			dlc->err_proc, dlc->err_ptr);
	    }

	}


      return led;    /* unchanged pointer to led */
    }  
  else 
    {
      dlc_delete(dlc,events_class,led);     /* remove event from old class */
      led=dlc_store(dlc,events_new_class,ued,r);  /* put it in new class,  */
                                                  /* get new led           */

      return led;   /* pointer to new led */
    }
}   /* end -- dlc_knownchange */



/*--------------------------------------------------------------------------*/
dlc_event *dlc_unknownchange(dlc_global *dlc, dlc_event *led, void *ued, 
			     double r)

/* - move the event described by led/ued whose NEW reactivity is r to 
     the appropriate class
   - update event's reactivity, class reactivity, total reactivity
   Called by: not used in this version
   Calls    : dlc_getclass, dlc_knownchange
*/
     
{
  return dlc_knownchange(dlc, dlc_getclass(dlc, led->r), led, ued, r);
}
