/*
   
   THE LOGCLASS PACKAGE

   User interface, type declarations and function headers for the 
   dual-logarithmic-classes code.

   Idea and first coding      : Thomas Fricke
   Beta testing               : Dietmar Wendt
   Polishing & Documentation  : Hans Ekkehard Plesser, D. Wendt

   Version                    : 1.3
   Last modification          : October 24, 2002

   Documentation & References : see MarkovClasses.pdf

*/

/*

  Copyright (C) 1995, 2002 Thomas Fricke, Dietmar Wendt,
                Hans Ekkehard Plesser <hans.plesser@nlh.no>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/* ========================================================================= */
/*                                                                           */
/*                                  Contents                                 */
/*                                                                           */
/* ========================================================================= */

/* This file consists of four parts:

   - Includes and definition of constants:
       contains all #include directives required by logclass.*
       and defines all preprocessor constants

   - Public type definitions:
       dlc_event
       dlc_class
       dlc_global
     which the user needs to define corresponding variables

   - Public function declarations:
       dlc_init
       dlc_clear
       dlc_enter
       dlc_rand
       dlc_safechange
       dlc_resum
     which constitute the user interface of the logclass package

   - Private function declarations [in dlc.1.3.c]:
       dlc_delete
       dlc_do_absolutely_nothing
       dlc_error
       dlc_getclass
       dlc_reorg
       dlc_store
       dlc_knownchange
       dlc_unknownchange
      which are internal to logclass and should not be called by the user
*/


/* ========================================================================= */
/*                                                                           */
/*                 Includes and definition of constants                      */
/*                                                                           */
/* ========================================================================= */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <errno.h>
#include <limits.h>   /* compiling with ULTRIX gcc results in warning that */
#include <float.h>    /* FLT_MAX has been redefined by float.h, but this   */
                      /* does not seem to cause problems                   */

#ifndef NAME_MAX
#define NAME_MAX 200
#endif

#define DLC_EXTRA_BOTTOM 0     /* no. of surplus classes generated at both */
#define DLC_EXTRA_TOP    2     /* ends of range                            */


/* ========================================================================= */
/*                                                                           */
/*                        Public type definitions                            */
/*                                                                           */
/* ========================================================================= */

typedef struct DLC_EVENT{ void   *ued; /* pointer to user's event descriptor */
			  double r;    /* reactivity of the event            */
			} dlc_event;

/* Type of the LOGCLASS' event descriptors (leds), representing individual
   events and being the elementary data of the LOGCLASS package (elements
   of arrays F[] in Fricke&Wendt).

   Pointer to user's event descriptor is void *, so it may be cast to any
   pointer type and each user may choose her/his own type of ued.

   The user-program needs to read element ued; it may read element r, but it
   MUST NOT assign to any element; the entries are maintained by the
   LOGCLASS functions, so please KEEP YOUR HANDS OFF !!!

   For details, see logclass.doc.*
*/

/*--------------------------------------------------------------------------*/

typedef struct DLC_CLASS{ dlc_event         *bot, *top;
			  double            r, r_max, eps;
			  struct DLC_CLASS  *next;
			} dlc_class;

/* Type of class descriptors, each describing one class, see Fricke&Wendt.

   A class L[z] (dual-logarithmic class) is a set of events e whose
   reactivities Q(e) yield to the inequality
		 1/2 2^z <= Q(e) < 2^z
   and is represented as an array F[] of events.

   Elements:
     - bot  : pointer to array F[], i.e. its first element 
     - top  : pointer to the first led "behind" this array, i.e. top does not 
              point to a valid led, but serves as upper limit
     - r    : reactivity of the class
     - r_max: highest reactivity the class reached until it was empty
     - eps  : lowest reactivity in this class treated as non-zero, calculated
              from the maximal round-off-error for this class
     - next : links each class to its successor in a list, which is 
              (imperfectly) sorted by class reactivity (descending order)

   Direct contact to these class descriptors should not be necessary, 
   so please KEEP YOUR HANDS OFF !!!
*/


/*--------------------------------------------------------------------------*/

typedef struct DLC_GLOBAL { dlc_class  *cbeg, *cend, *first;
			    int        min_class, num_classes;
			    size_t     max_events;
			    double     r, eps;
			    long int   number_of_reorgs;
			    void       (*event_moved)(dlc_event *);
			    char       err_file[NAME_MAX];
			    void       (*err_proc)();
                            void       *err_ptr;
			  } dlc_global;

/* The global data structure of LOGCLASS containing all information required
   by LOGCLASS' functions to operate.  

   The user-program has to allocate memory for and provide a pointer to this
   structure, e.g. defining
     dlc_global global_data, *global_data_ptr = &global_data;
   this pointer is passed to all LOGCLASS functions as first argument.

   All entries are set by function dlc_init and maintained by LOGCLASS
   throughout program execution.  They are

   cbeg   : pointer to descriptor of the class with lowest reactivity, i.e.
            "beginning" of the array of class descriptors
   cend   : pointer to descriptor of an empty class, marking the upper
            limit of the array, i.e. valid pointers to classes are given
	    by
	       cbeg[n] < cend, 0 <= n ( < num_classes )
   first  : pointer to first element in linked list of class descriptors,
            i.e. class where dlc_rand starts search
   min_class  : index of lowest class
   num_classes: number of class descriptors allocated, i.e. number of elements
                in array beginning at cbeg EXCLUDING dummy class cend; 
		event reactivities < 2^(min_class-1) are treated as zero,
		reactivities >= 2^(min_class+num_classes-1) cause an overflow
   max_events : number of LOGCLASS event descriptors (leds) allocated, i.e.
                maximum number of events that can be handled
   r          : total reactivity of the system, 1/r is mean of the
                first passage time distribution of the Markoff-events
   eps        : lowest reactivity not treated as zero, i.e. 2^(min_class-1)
   number_of_reorgs: counts how often memory has been reorganized since
                     initialization 
   *event_moved    : pointer to function defined in user-program to update
                     the link ued->led
   err_file        : name of file for error messages
   *err_proc       : pointer to function defined in user-program to "wrap
                     things up" before exiting due to error in LOGCLASS
   *err_ptr        : in case of error, err_proc is called with err_ptr as
                     argument, so the user can pass a reference to his/her
                     data to the "wrap up routine"; if you have no such
                     pointer, pass NULL.

   The element r is used by the user program to determine the time step 
   according to Gillespie.
   The element number_of_reorgs may be read to monitor "memory performance".

   Any other contact to the global data should not be necessary, 
   so please KEEP YOUR HANDS OFF !!!
*/



/* ========================================================================= */
/*                                                                           */
/*                      Public function declarations                         */
/*                                                                           */
/* ========================================================================= */
void dlc_init(dlc_global *dlc, size_t num_leds, double min_reac, 
	      double max_reac, long int rand_seed, void (*event_moved)(), 
	      char err_file[], void (*err_proc)(), void *err_ptr);

/* Task: 
     - allocates memory for class and LOGCLASS event descriptors
     - initializes elements of dlc_global data structure
     - sets seed for random generator

   Arguments: 
     - dlc        : pointer to global data structure
     - num_leds   : number of LOGCLASS event descriptors to allocate
     - min_reac   : estimated lowest reactivity
     - max_reac   : estimated highest reactivity
     - rand_seed  : random seed, use 0L to take clock as seed
     - event_moved: (pointer to) function updating link ued->led
     - err_file   : string with name of file for error messages
     - err_proc   : (pointer to) function "wrapping up" after an error
     - err_ptr    : argument to be passed to err_proc 

   Exits with error message if
     - min_reac <= 0
     - range of reactivities to large
     - to few leds have been allocated per class
     - there was not enough memory available

   Called by:
     user-program

   Calls:
     dlc_error

   Remark: 
     MUST be called before any other LOGCLASS routine !
*/



/*--------------------------------------------------------------------------*/
void dlc_clear(dlc_global *dlc);

/* Task: 
     - returns memory allocated by dlc_init() to operating system

   Arguments: 
     - dlc : pointer to global data structure
     
   Called by:
     user-program

   Remarks: 
     - Should be called at the end of simulation
*/



/*--------------------------------------------------------------------------*/
dlc_event *dlc_enter(dlc_global *dlc, void *ued, double r);

/* Task:
     - assigns an led to the event represented by ued and places it in a
       class according to its reactivity r; sets up link led->ued and
       stores reactivity in led

   Arguments:
     - dlc : pointer to global data struture
     - ued : pointer to events descriptor in user-programs data structure
     - r   : reactivity of the event

   Returns:
     - pointer to led

   Exits with error message and error code 1, if
     - frexp() sets errno
     - r < 0
     - reactivity is out of range of classes
     - dlc or ued are NULL
     - all leds allocated have been assigned already
     - memory reorganization failed

  Called by:
    user-program, dlc_safechange

  Calls:
    dlc_getclass, dlc_store

  Remarks:
     - an event's ued must exist before it can be assigned a led
     - events MUST be assigned a led before they can be selected
     - dlc_enter should be called by the user-program only once for each event
*/



/*--------------------------------------------------------------------------*/
dlc_event *dlc_rand(dlc_global *dlc);

/* Task:
     - select one event at random as described by Fricke&Wendt

   Arguments:
     - dlc         : pointer to the global data structure
     - re_class_ptr: pointer to return value, see below

   Returns:
     - dlc_rand    : pointer to the led of the event selected
     - re_class_ptr: pointer to pointer to the class descriptor of
                     the class from which the event was selected,
		     provided for use by dlc_knownchange

   Exits with error message and error code 1, if
     - a non-existing class was selected
     - the class selected does not contain any events
     - no class was selected at all

   Called by:
     user-program

   Calls:
     dlc_error

   Remarks:
     - as a side effect, dlc_rand exchanges adjacent elements of the
       linked list of class descriptors if they are no longer in descen-
       ding order of class reactivity, so the list remains partially
       sorted
*/



/*--------------------------------------------------------------------------*/
dlc_event *dlc_safechange(dlc_global *dlc, dlc_event *led, void *ued, 
			  double r);
/* Task:
     - move or enter the event described by led/ued whose NEW reactivity is r
       to the appropriate class
     - update event's reactivity entry in led, class reactivities of all
       classes involved and total reactivity
     - reorganize memory if necessary

   Arguments:
     - dlc      : pointer to global data structure
     - led, ued : pointers to events led/ued, respectively
     - r        : NEW reactivity of the event

  Returns:
     - NEW pointer to the event's led, NULL if r==0

   Exits with error message, if
     - serious round-off errors are detected
     - frexp() sets errno
     - r < 0
     - reactivity is out of range of classes
     - dlc or ued are NULL
     - all leds allocated have been assigned already
     - memory reorganization failed

   Called by:
     user-program

   Calls:
     dlc_enter, dlc_knownchange

   Remarks:
     - led == NULL will work correctly
     - does NOT update link ued -> led, since it is faster if the user program
       does it, e.g. by
         ued->led_ptr = dlc_safechange(dlc, led, ued, r);
*/



/*--------------------------------------------------------------------------*/
int  dlc_resum(dlc_global *dlc);
/* Task:
     - resummarize the reactivities of all not-empty classes
     - set the reactivity of empty classes to 0
     - resummarize dlc->r with the new class-reactivities

   Arguments:
     dlc : pointer to global data structure

   Returns:
     number of wrong old sums (class->r for all classes, dlc->r)

   Called by:
     user-program

   Calls:

   Remarks:
*/



