/****************************************************************************/
/*                                                                          */
/*     MARKOV CLASSES: LOGCLASS --- SAMPLE PROGRAM                          */ 
/*                                                                          */
/****************************************************************************/

/*

  Copyright (C) 2002 Hans Ekkehard Plesser <hans.plesser@nlh.no>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/* Simulation of Simplified Lotka-Volterra model    

        A     -> 2A   (  birth  , rate constant kA)
        A + B -> 2B   (predation, rate constant kB)
        B     -> 0    (  death  , rate constant kD)

   A and B diffusion in 2 dimensions, rate dA, dB
   
   At certain intervals, some global information is logged to logfile.
   (time, number of steps, total rate, total A, B numbers, number of
   reorganizations)

   At other intervals, the total state of the system is
   dumped into files dumpfile.0000, dumpfile.0001 ...  The left column
   is #A, the second #B, cells are dumped from left to right, top to
   bottom.

   Invoke as

   dlc_demo paramfile outfile

   Logging is then to outfile.log, errors go to outfile.err, and dumps
   go to outfile.0000, outfile.0001, ...  paramfile is a file adhering to 
   precisely the following structure (------- lines not included)

--------------------------
L     : 50
Tmax  : 100
dtdump: 0.1
dtlog : 0.02
seed  : 1234567
dA    : 0.1
dB    : 0.2
kB    : 1
kP    : 0.1
kD    : 1
nA(0) : 10
nB(0) : 10
--------------------------

   If seed is 0, the seed will be taken from the clock (resolution ca 1 sec).
   nA(0), nB(0) are the initial population numbers per cell.

*/

/* include files */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include <rand55.h>

#include "dlc.h"     /* the logarithmic classes */

/****************************************************************************/
/*                                                                          */
/*                                 Parameters                               */
/*                                                                          */
/****************************************************************************/

#define MEMORY_FACTOR    3     /* extra leds             */
#define MAX_FN         255     /* max filename length    */
#define MIN_RATE         1e-5  /* rate limits for setup  */       
#define MAX_RATE         1e15  /* of class array         */

/****************************************************************************/
/*                                                                          */
/*                             Type Declarations                            */
/*                                                                          */
/****************************************************************************/

typedef struct CELL_STATE     /* this is the type for the ued */
{ 
  unsigned long      nA;          /* number of A in cell */
  unsigned long      nB;          /* number of B in cell */
  double             rB,          /* birth rate for cell */ 
                     rP,          /* pred rate for cell  */
                     rD,          /* death rate for cell */ 
                     sA, sB;      /* diff rates for cell */
  double             rate;        /* rB + rP + rD + sA + sB */

  dlc_event          *led;       /* ptr to cell's DLC event descriptor */
} cell_state;

/* global simulation data structure */
typedef struct 
{
  unsigned long L;            /* system length: L cells */
  unsigned long num_cells;    /* number of cells = L^2  */

  unsigned long tot_A;        /* total number of A      */
  unsigned long tot_B;        /* total number of B      */
  double        t;            /* time                   */
  double        Tmax;         /* time limit             */
  unsigned long long nstep;   /* step number            */

  double        dA, dB,       /* diffusion rate constants */
                kB, kP, kD;   /* reaction rate constants  */
  double        nA0, nB0;     /* initial A, B count per cell */ 

  double        dt_dump;      /* time interval between dumps */
  double        next_dump;    /* time of next dump           */
  unsigned long ndump;        /* dump number                 */
  double        dt_log;       /* time interval between logs  */
  double        next_log;     /* time of next log            */

  char          params[MAX_FN];    /* parameter file name   */
  char          logname[MAX_FN];   /* log file name         */
  char          errname[MAX_FN];   /* error file            */
  char          dumpstub[MAX_FN];  /* dump file stub        */

  unsigned long rand_seed;         /* random seed           */

  cell_state    *first_cell;       /* limits of cell list   */
  cell_state    *last_cell;

  FILE          *logfile;          

} sim_data;


/****************************************************************************/
/*                                                                          */
/*                            Function Prototypes                           */
/*                                                                          */
/****************************************************************************/

/* handle event in one cell chosen by dlc_rand */
void handle_cell(dlc_global*, sim_data*, cell_state*);

/* actually execute one event */
cell_state *execute_event(sim_data *sd, cell_state *cell);

/* compute cell's rate */
double compute_rate(sim_data *sd, cell_state *cell);

/* helper functions */
void init_system(sim_data *sd);
void log_status(dlc_global *classes, sim_data *sd);
void dump_status(sim_data *sd);

/* functions to be passed to dlc_init for event and error handling */
void  ued2led_update(dlc_event*);
void  save_on_error();

/****************************************************************************/
/*                                                                          */
/*                               Main Program                               */
/*                                                                          */
/****************************************************************************/

int main(int argc, char *argv[])
{ 
  /* memory, pointer for your global data */
  sim_data   s, *sd=&s;                

  /* memory, pointer for your discrete classes' global data */
  dlc_global  cls, *classes = &cls;     

  /* local variables necessary for discrete class handling  */
  cell_state *cell;       /* pointer to ued                  */
  dlc_event   *ce;        /* pointer to selected event       */

  /* your local variables */

  /* read out command line arguments */
  if ( argc != 3 )
    {
      printf("\n dlc_demo requires two arguments.\n");
      return -1;
    }
  strcpy(sd->params, argv[1]);    /* rates constants, time limit */
  strcpy(sd->logname, argv[2]);
  strcat(sd->logname, ".log");
  strcpy(sd->errname, argv[2]);
  strcat(sd->errname, ".err");
  strcpy(sd->dumpstub, argv[2]);

  /* initialize your data structures */
  init_system(sd);

  /* initialize logarithmic classes */
  dlc_init(classes, MEMORY_FACTOR * sd->num_cells, MIN_RATE, MAX_RATE,
	   sd->rand_seed, ued2led_update, sd->errname, save_on_error, sd);

  /* enter all cells in appropriate classes */
  for ( cell = sd->first_cell ; cell < sd->last_cell ; cell++ )
    cell->led = dlc_enter(classes, cell, cell->rate);

  dump_status(sd);
  sd->logfile = fopen(sd->logname, "w");
  assert(sd->logfile);
  log_status(classes, sd);

  while ( sd->t < sd->Tmax 
          && (classes->r > 0) )             /* system has not ``died'' */
    {
      /* draw time step with exponential distribution */
      sd->t += exp_rand55() / classes->r;

      /* dump prior to next step if requested */
      if ( sd->t > sd->next_dump )
	{
	  dump_status(sd);
	  sd->next_dump += sd->dt_dump;
	}
      if ( sd->t > sd->next_log  )
	{
	  log_status(classes, sd);
	  sd->next_log += sd->dt_log;
	}

      /* let LOGCLASS choose the "active" cell */
      ce   = dlc_rand(classes);
      cell = (cell_state *) ce->ued;      /* get pertaining ued  */

      /* execute an event in the cell selected */
      handle_cell(classes, sd, cell);
      
      sd->nstep++;

     }    /* end while */

  log_status(classes, sd);
  dump_status(sd);

  fclose(sd->logfile);
  dlc_clear(classes);    /* clean up DISCLASS data structures */

  /* do whatever you want */

  return 0;

}   /* end main */


/***************************************************************************/

void handle_cell(dlc_global *classes, sim_data *sd, cell_state *cell)

{ 
  cell_state *target;

  /* find and perform event, target returns target cell
     of diffusion step, null otherwise                  */
  target = execute_event(sd, cell);

  /* recompute rate and reassign cell */
  compute_rate(sd, cell);
  cell->led = dlc_safechange(classes, cell->led, cell, cell->rate);

  /* if we have a target, same for target */
  if ( target )
    {
      compute_rate(sd, target);
      target->led = dlc_safechange(classes, target->led, target, target->rate);
    }

  return;

} /* end -- handle_cell */

/***************************************************************************/

cell_state *execute_event(sim_data *sd, cell_state *cell)

{
  cell_state *target;
  double r;

  /* check what to do */
  r = cell->rate * drand55();

  /* diffusion processes have highest rates, check them first */
  if ( r < cell->sA )   /* A diffusion */
    {
      /* choose target */
      long index = cell - sd->first_cell;

      switch ( lrand55(4) ) {
      case 0:  /* left */
	index -= 1;
	break;
      case 1:  /* right */
	index += 1;
	break;
      case 2:  /* up    */
	index -= sd->L;
	break;
      case 3:  /* down */
	index += sd->L;
	break;
      default:
	assert(0);
	break;
      }

      /* helical boundary conditions on cell grid */
      if ( index < 0 )              index += sd->num_cells;
      if ( index >= sd->num_cells ) index -= sd->num_cells;
      target = sd->first_cell + index;

      cell->nA--;
      target->nA++;

      /* no change of total particle number */

      return target;
    }

  r -= cell->sA;

  if ( r < cell->sB )   /* B diffusion */
    {
      /* choose target */
      long index = cell - sd->first_cell;

      switch ( lrand55(4) ) {
      case 0:  /* left */
	index -= 1;
	break;
      case 1:  /* right */
	index += 1;
	break;
      case 2:  /* up    */
	index -= sd->L;
	break;
      case 3:  /* down */
	index += sd->L;
	break;
      default:
	assert(0);
	break;
      }

      /* helical boundary conditions on cell grid */
      if ( index < 0 )              index += sd->num_cells;
      if ( index >= sd->num_cells ) index -= sd->num_cells;
      target = sd->first_cell + index;

      cell->nB--;
      target->nB++;

      /* no change of total particle number */

      return target;
    }

  r -= cell->sB;

  /* no diffusion, so no other cell affected */
  target = NULL;

  if ( r < cell->rB )   /* A birth process */
    {
      cell->nA++;
      sd->tot_A++;
      return target;
    }

  r -= cell->rB;

  if ( r < cell->rP )   /* predation */
    {
      cell->nA--;
      cell->nB++;
      sd->tot_A--;
      sd->tot_B++;
      return target;
    }

  /* default event if all else fails, but safeguard against numerical errors */
  if (cell->nB > 0)
    { 
      cell->nB--;
      sd->tot_B--;
      return target;
    }

  /* just for safety */
  return NULL;

}  /* execute_event */

/****************************************************************************/

double compute_rate(sim_data *sd, cell_state *cell)

{
  /* recompute rates */
  cell->sA = sd->dA * cell->nA;
  cell->sB = sd->dB * cell->nB;
  cell->rB = sd->kB * cell->nA;
  cell->rP = sd->kP * cell->nA * cell->nB;
  cell->rD = sd->kD * cell->nB;
  cell->rate = cell->sA + cell->sB + cell->rB + cell->rP + cell->rD;

  return cell->rate;
}


/***************************************************************************/

void ued2led_update(dlc_event *new_led)


{ 
     /* update ued pointer or led */

    ( (cell_state *) new_led->ued )->led = new_led; 
}


/***************************************************************************/

void save_on_error(sim_data *sd)

{
  /* preserve final state when something went wrong */
  dump_status(sd);
}


/****************************************************************************/

void init_system(sim_data *sd)

{
  FILE *dfile;
  cell_state *cell;

  dfile = fopen(sd->params, "r");
  assert(dfile);

  fscanf(dfile, "L     : %lu\n", &(sd->L)); 
  fscanf(dfile, "Tmax  : %le\n", &(sd->Tmax)); 
  fscanf(dfile, "dtdump: %le\n", &(sd->dt_dump)); 
  fscanf(dfile, "dtlog : %le\n", &(sd->dt_log)); 
  fscanf(dfile, "seed  : %lu\n", &(sd->rand_seed)); 
  fscanf(dfile, "dA    : %le\n", &(sd->dA)); 
  fscanf(dfile, "dB    : %le\n", &(sd->dB)); 
  fscanf(dfile, "kB    : %le\n", &(sd->kB)); 
  fscanf(dfile, "kP    : %le\n", &(sd->kP)); 
  fscanf(dfile, "kD    : %le\n", &(sd->kD)); 
  fscanf(dfile, "nA(0) : %le\n", &(sd->nA0)); 
  fscanf(dfile, "nB(0) : %le\n", &(sd->nB0)); 

  fclose(dfile);

  /* allocate cell memory */
  sd->num_cells = sd->L * sd->L;
  sd->first_cell = calloc(sd->num_cells+1, sizeof(cell_state));
  assert(sd->first_cell);
  sd->last_cell = sd->first_cell + sd->num_cells;

  /* set initial condition */
  for ( cell = sd->first_cell ; cell < sd->last_cell ; cell++ )
    {
      cell->nA = sd->nA0;
      cell->nB = sd->nB0;
      compute_rate(sd, cell);
    }
  sd->tot_A = sd->num_cells*sd->nA0;
  sd->tot_B = sd->num_cells*sd->nB0;

  /* init values */
  sd->t = 0.0;
  sd->next_dump = sd->dt_dump;
  sd->ndump = 0;
  sd->next_log  = sd->dt_log;
  sd->nstep     = 0ULL;

}

/****************************************************************************/

void log_status(dlc_global *classes, sim_data *sd)

{
  fprintf(sd->logfile, "%20llu %12.6e  %12.6e  %12lu  %12lu  %6u\n",
	  sd->nstep, sd->t, classes->r, sd->tot_A, sd->tot_B, 
	  classes->number_of_reorgs);
}

/****************************************************************************/

void dump_status(sim_data *sd) 

{
  char dumpname[MAX_FN];
  FILE *dumpfile;
  cell_state* cell;

  sprintf(dumpname, "%s.%04lu", sd->dumpstub, sd->ndump);
  dumpfile = fopen(dumpname, "w");

  if ( dumpfile )
    {
      for ( cell = sd->first_cell ; cell < sd->last_cell ; cell++ )
	fprintf(dumpfile, "%12lu %12lu\n", cell->nA, cell->nB);

      fclose(dumpfile);
      sd->ndump++;
    }
}


/****************************************************************************/
/*                                                                          */
/*                             End of Program                               */
/*                                                                          */
/****************************************************************************/
